#include "nawiglcanvas.h"
#include "nawiframe.h"
#include "world.h"
#include "camera.h"
#include <wx/dcclient.h>

BEGIN_EVENT_TABLE(NawiGLCanvas, wxGLCanvas)
	EVT_SIZE		(NawiGLCanvas::onSize)
	EVT_PAINT		(NawiGLCanvas::onPaint)
	EVT_IDLE		(NawiGLCanvas::onIdle)
	EVT_ERASE_BACKGROUND	(NawiGLCanvas::onEraseBackground)
	EVT_KEY_UP		(NawiGLCanvas::onKeyUp)
	EVT_KEY_DOWN		(NawiGLCanvas::onKeyDown)
END_EVENT_TABLE()


NawiGLCanvas::NawiGLCanvas(NawiFrame *parentWindow)
	:
	wxGLCanvas( parentWindow, wxID_ANY, NULL, wxDefaultPosition, parentWindow->GetClientSize() ),
	parentWindow(parentWindow),
	camera(NULL),
	wireFrame(false)
{
	glContext = new wxGLContext(this);
	SetFocus();
}

NawiGLCanvas::~NawiGLCanvas()
{
	if( glContext )
		delete glContext;
	if( camera )
		delete camera;
}

void NawiGLCanvas::onSize(wxSizeEvent &event)
{
	if( camera )
		camera->setViewport( GetSize() );
}

void NawiGLCanvas::onEraseBackground( wxEraseEvent& WXUNUSED(event) )
{
	// empty function,to avoid flashing.
}

void NawiGLCanvas::onKeyUp(wxKeyEvent &e)
{
	if( e.m_controlDown )
	{
		switch( e.m_keyCode )
		{
			case 'W':
				wireFrame = !wireFrame;
				break;
		}
	}
	else if( e.m_altDown )
	{
		switch( e.m_keyCode )
		{
		case 'R':
			camera->focal = Focal( camera->viewport.getDiagonal() );
			camera->updateProjection();
			break;
		}
	}
	else
	{
		switch( e.m_keyCode )
		{
			case 'R':
				camera->reset();
				break;
		}
	}
	e.Skip();
	Refresh();
}

void NawiGLCanvas::onKeyDown(wxKeyEvent &e)
{
	World *world = parentWindow->world;

	if( e.m_altDown )
	{
		switch( e.m_keyCode )
		{
		case '=':
			camera->focal.length *= 1.01;
			camera->updateProjection();
			break;
		case '-':
			camera->focal.length *= 0.99;
			camera->updateProjection();
			break;
		}
	}
	else if( e.m_controlDown )
	{
	}
	else if( e.m_shiftDown )
	{
		switch( e.m_keyCode )
		{
		// Incline to the right
		case 'D':
			camera->orientation *= Quaternion( camera->viewMatrix.getRow(2), -0.01 );
			break;
		// Incline to the left
		case 'Q':
			camera->orientation *= Quaternion( camera->viewMatrix.getRow(2), 0.01 );
			break;
		}
	}
	else
	{
		Quaternion rot;
		double scale = camera->altitude / world->radius;
		Vector cameraToWorldCenter = camera->getPosition() - world->position;
		switch( e.m_keyCode )
		{
		// Rotate up
		case 'Z':
			camera->orientation = Quaternion( Vector(1., 0., 0.), -0.01 ) * camera->orientation;
			break;
		// Rotate down
		case 'W':
			camera->orientation = Quaternion( Vector(1., 0., 0.), 0.01 ) * camera->orientation;
			break;
		// Rotate to the right
		case 'D':
			camera->orientation = Quaternion( Vector(0., 1., 0.), 0.01 ) * camera->orientation;
			break;
		// Rotate to the left
		case 'Q':
			camera->orientation = Quaternion( Vector(0., 1., 0.), -0.01 ) * camera->orientation;
			break;
		// Pan left
		case WXK_LEFT:
			rot = Quaternion( camera->viewMatrix.getRow(1), 0.01 * scale );
			camera->setPosition( rot * camera->getPosition() );
			camera->orientation *= rot;
		break;
		// Pan down
		case WXK_DOWN:
			rot = Quaternion( camera->viewMatrix.getRow(0), -0.01 * scale );
			camera->setPosition( rot * camera->getPosition() );
			camera->orientation *= rot;
			break;
		// Pan right
		case WXK_RIGHT:
			rot = Quaternion( camera->viewMatrix.getRow(1), -0.01 * scale );
			camera->setPosition( rot * camera->getPosition() );
			camera->orientation = camera->orientation * rot;
			break;
		// Pan up
		case WXK_UP:
			rot = Quaternion( camera->viewMatrix.getRow(0), 0.01 * scale );
			camera->setPosition( rot * camera->getPosition() );
			camera->orientation = camera->orientation * rot;
			break;
		// Go down to the ground, or up to the sky
		// Interval gets smaller near the ground
		case WXK_PAGEDOWN:
			camera->setPosition(
				camera->getPosition() + cameraToWorldCenter * (
					camera->altitude
					/ ( camera->altitude + world->radius )
					* -0.1
					)
				);
			break;
		case WXK_PAGEUP:
			camera->setPosition(
				camera->getPosition() + cameraToWorldCenter * (
					camera->altitude
					/ ( camera->altitude + world->radius )
					* 0.1 
					)
				);
			break;
		// Zoom in or out
		// Interval gets smaller near the ground
		case 'S':
		case '=':
			camera->setPosition(
				camera->getPosition()
				- camera->viewMatrix.getRow(2) * ( scale * 400000. )
				);
			break;
		case 'X':
		case '-':
			camera->setPosition(
				camera->getPosition()
				+ camera->viewMatrix.getRow(2) * ( scale * 400000. )
				);
			break;
		case WXK_ESCAPE:
			parentWindow->Close();
			break;
		}
	}
	e.Skip();
	Refresh();
}

void NawiGLCanvas::onPaint(wxPaintEvent &event)
{
	SetCurrent(*glContext);
	wxPaintDC(this);
	parentWindow->render();
	glFlush();
	SwapBuffers();
}

void NawiGLCanvas::onIdle(wxIdleEvent &event)
{
	if( parentWindow->world->updatedSinceLastRender )
	{
		SetCurrent(*glContext);
		wxClientDC(this);
		parentWindow->render();
		glFlush();
		SwapBuffers();
	}
}
